/* fEopen.c -- Copyright 1988,1992 Liam R. E. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* $Id: fEopen.c,v 1.2 92/04/29 13:39:19 lee Exp $
 * Error checking versions of fopen() and fclose(), taken from Liam Quin's
 * unreleased error handling library and stripped down somewhat.
 */

#include "globals.h"
#include "emalloc.h"
#include "error.h"

#include <stdio.h>
#include <malloc.h>	/* This declares malloc(), realloc() etc... */

#include <errno.h>
#ifdef SYSV
# include <string.h>
#else
# include <strings.h>
# include <sys/types.h> /* for fcntl.h */
# include <sys/stat.h> /* for fcntl.h */
#endif

#include <fcntl.h> /* for O_RDONLY etc */

/** error-checking fopen and fclose...
 **
 ** Contents:
 **
 **
 ** FILE *fEopen(int Severity; char *Name; char *What; char *Mode);
 **	This opens the Named file in the given mode; otherwise, it
 **	calls Error() with the appropriate severity.  "What" is a short
 **	description of the file -- e.g. "boys' names".  Error
 **	messages include this, for example
 **	prog: fatal error: couldn't open "name" (boys' names): file not found
 **
 ** void fEclose(int severity; FILE *fp; char *Name; char *What);
 **/

char *FindMissingPart();
static void Diagnose();

int
Eopen(Severity, Name, What, Flags, Modes)
    int Severity;
    char *Name;
    char *What;
    int Flags;
    int Modes;
{
    int Result;

    errno = 0;
    Result = open(Name, Flags, Modes);

    if (!Name) {
	char *Format = "couldn't open \"%s\" (%s)";

	if (What && *What) {
	    Error(Severity|E_INTERNAL|E_BUG,
		Format,
		"[NULL name]",
		What
	    );
	} else {
	    Error(Severity|E_INTERNAL|E_BUG,
		"Eopen(%d, 0, 0, %d, %d) illegal",
		Severity,
		Flags,
		Modes
	    );
	}
	return -1;
    }

    if (IsDir(Name)) {
	Error(Severity,
	    "file \"%s\" (%s) is a directory (not opened)",
	    Name,
	    What
	);
	return -1;
    }

    if (Result >= 0) {
	return Result;
    } else {
	Diagnose(
	    Severity,
	    Name,
	    What,
	    errno,
	    Flags
	);
	return -1;
    }
}

static int
StdioModeToFlags(Mode)
    char *Mode;
{
    int Flags = O_RDONLY;
    register char *p;

    for (p = Mode; *p; p++) {
	switch (*p) {
	case 'w': case 'W':
	    Flags |= O_WRONLY|O_CREAT|O_TRUNC;
	    break;
	case 'a':
	case 'A':
	    Flags |= O_APPEND;
	    Flags &= ~O_TRUNC;
	    break;
	case '+':
	    Flags |= O_RDWR|O_CREAT;
	    Flags &= ~O_TRUNC;
	    break;
	}
    }
    return Flags;
}

FILE *
fEopen(Severity, Name, What, Mode)
    int Severity;
    char *Name;
    char *What;
    char *Mode;
{
    FILE *Result;
    char *Format = "couldn't open \"%s\" (%s)";

    if (!Name) {
	if (What && *What) {
	    Error(Severity|E_INTERNAL, Format, "[NULL name]", What);
	} else {
	    Error(Severity|E_INTERNAL, "fEopen(%d, 0, 0, %s) illegal",
			Severity, (Mode ? ((*Mode)? Mode : "\"\"" ) : "0"));
	}
	return (FILE *) 0;
    }

    if (!Mode || !*Mode) {
	if (!What || !*What) {
	    Error(Severity|E_INTERNAL,
		 "fEopen: couldn't open \"%s\" (%s), since no Mode was given",
		 Name, What);
	} else {
	    Error(Severity|E_INTERNAL,
		 "fEopen: couldn't open \"%s\", since no Mode was given",
		 Name);
	}
	return (FILE *) 0;
    }

    if (IsDir(Name)) {
	Error(Severity,
	    "file \"%s\" (%s) is a directory (not opened)",
	    Name,
	    What
	);
	return (FILE *) 0;
    }

    errno = 0;
    if ((Result = fopen(Name, Mode)) != (FILE *) 0) {
	return Result;
    } else {
	Diagnose(
	    Severity,
	    Name,
	    What,
	    errno,
	    StdioModeToFlags(Mode)
	);
    }
    return (FILE *) 0;
}

static void
Diagnose(Severity, Name, What, Errno, Flags)
    int Severity;
    char *Name;
    char *What;
    int Errno;
    int Flags;
{
    char *Problem;

    switch (Errno) {
    case EFAULT:
	Error(Severity|E_BUG,
	    "open called with invalid file-name (EFAULT)"
	);
	/* NOTE: don't print "what" out in this case lest it's broken too */
	break;

    case E2BIG:
#ifdef ENAMETOOLONG
    case ENAMETOOLONG:
#endif
	Error(Severity|E_SYS,
	    "Can't open file \"%1024.1024s...\" (%s)",
	    Name,
	    What
	);
	break;
    case EPERM:
    case ENOENT:
    case EACCES:
    case ENOTDIR:

	if ((Problem = FindMissingPart(Name)) != (char *) NULL) {
	    if (STREQ(Problem, Name)) {
		Error(Severity,
		    "\"%s\" (%s) not opened",
		    Name,
		    What
		);
	    } else {
		Error(Severity|E_MULTILINE,
		    "\"%s\" (%s) not opened",
		    Name,
		    What
		);

		Severity &= (~E_SYS);

		if (IsFile(Problem)) {
		    Error(Severity|E_LASTLINE,
			"\"%s\" is a file, not a directory",
			Problem
		    );
		} else {
		    Error(Severity|E_LASTLINE,
			"\"%s\" is not a directory",
			Problem
		    );
		}
	    }
	    (void) free(Problem); /* NOTE: not efree() */
	    break;
	}
    }

    /* if we get here, there was another problem opening the file */
    Error(Severity|E_SYS,
	"can't open file \"%s\" (%s) Flags %d",
	Name,
	What,
	Flags
    );
}

static char *
NextSlash(String)
    register char *String;
{
    for (; *String; String++) {
	if (*String == '/') {
	    break;
	}
    }
    return String;
}

char *
FindMissingPart(BadPath)
    char *BadPath;
{
    int e = errno;
    char *Path;
    register char *End;

    /* We must look to find out which component of the path is missing.
     * Note that the parameter BadPath might be read-only...
     */
    
    if (!BadPath || !*BadPath) return (char *) NULL;

    /* don't use emalloc, because we are called in the middle of
     * gnerating an error message, and we want the _real_ message
     * to be produced, not
     * "out of memory trying to produce a clearer error message" !
     */

    Path = malloc(strlen(BadPath) + 1);
    
    if (Path == (char *) NULL) {
	errno = e;
	return (char *) NULL;
    }

    /* Loop invariants:
     *
     * Precondition:
     *  
     * Invariant:
     *	*End == '/'
     * Postcondition:
     *  *End == '\0'
     */

    End = BadPath;
    while (*End) {
	int n;

	End = NextSlash(&End[1]);

	if (!*End) {
	    break;
	}

	n = End - BadPath;
	(void) strncpy(Path, BadPath, n);
	Path[n] = '\0';

	if (!IsDir(Path)) {
	    return Path;
	}
    }

    (void) strcpy(Path, BadPath);
    errno = e;
    return Path;
}

void
fEclose(Severity, fp, Name, What)
    int Severity;
    FILE *fp;
    char *Name;
    char *What;
{
    int Val;

    if (!Name) {
	if (fp == stdin) Name = "standard input";
	else if (fp == stdout) Name = "standard output";
	else if (fp == stderr) Name = "standard error";
    }

    if (!fp) {
	Error(Severity|E_INTERNAL, "fEclose %s [%s] with NULL fp forbidden",
	    (What ? What : "stream"), (Name ? Name : " (unnamed) "));
    } else if (!Name) {
	Error(Severity|E_INTERNAL, "fEclose %s with NULL Name forbidden",
		What ? What : "a file");
    }

    if ((Val = fclose(fp)) != 0) {
	Error(Severity|E_SYS, "%s: error closing %s",
					Name, What ? What : "file");
    }
}
